// Copyright (c) 2013 The Chromium OS Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef QUIPPER_STRING_
#define QUIPPER_STRING_

#ifndef HAS_GLOBAL_STRING
using std::string;
using std::stringstream;
#endif

#endif  // QUIPPER_STRING_
