// Copyright (c) 2013 The Chromium OS Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef QUIPPER_TEST_H_
#define QUIPPER_TEST_H_

#include <gtest/gtest.h>

#endif  // QUIPPER_TEST_H_
